#-------------------------------------------------------------------------------------#
# We-Bot Pi 超音波センサ(HC-SR04)プログラム(Ver.0 2019/07/31)
#-------------------------------------------------------------------------------------#
# ソフトウェアを利用するにあたっての注意点（必ず確認して下さい）
# ・本ソフトウェアは(株)ケイエスワイがデモ用に作成したソフトです。
# ・本ソフトウェアの利用にあたっては、全て利用者の自己責任で行って下さい。
# ・本ソフトウェアの改変、流用は全て利用者の責任において、自由に行って構いません。
# ・本ソフトウェアを使用して発生したあらゆる損害（金銭的損害、物理的損害等）は
#   全て利用者の責任において負担し、作者はいかなる損害も補償しません。
# ・本ソフトウェアの配布等に関しては、全て配布者の責任において行い、
#   作者は一切の責任を負わないものとします。
# ・作者は本ソフトウェアに関しての問い合わせ等には一切ご対応致しません。

#-------------------------------------------------------------------------------------#
# ライブラリ
#-------------------------------------------------------------------------------------#
import sys
import time
import RPi.GPIO as GPIO

#-------------------------------------------------------------------------------------#
# 定数定義
#-------------------------------------------------------------------------------------#
RETURN_VALUE_SUCCESS = 0        # 
RETURN_VALUE_ERROR = -1         # 

GPIO_SONIC_TRIG = 26            # 超音波センサ TRIGピン GPIO26(37ピン)
GPIO_SONIC_ECHO = 19            # 超音波センサ TRIGピン GPIO19(35ピン)

#-------------------------------------------------------------------------------------#
# 関数
#-------------------------------------------------------------------------------------#
def init_hcsr04(trig_pin=26, echo_pin=19, gpio_mode=GPIO.BCM):
    """
    超音波センサ関連のGPIOを初期化する
    :param trig_pin: トリガピン番号(default: 26)
    :param echo_pin: エコーピン番号(default: 19)
    :param gpio_mode: GPIOの指定方法(GPIO.BCMかGPIO.BOARD。default:GPIO.BCM)
    :return: なし
    """

    # GPIOの初期化
    GPIO.setwarnings(False)
    GPIO.setmode(gpio_mode)
    GPIO.setup(trig_pin, GPIO.OUT)
    GPIO.output(trig_pin, GPIO.LOW)
    GPIO.setup(echo_pin, GPIO.IN)


def measure_pulse_time(echo_pin=19, pin_value=GPIO.HIGH, timeout=0.25):
    """
    ECHOピンに入力されるパルスの幅を測定する
    :param echo_pin: ECHOのGPIO 番号(default:GPIO19)
    :param pin_value: ECHOピンのレベル指定(GPIO.HIGH or GPIO.LOW default:GPIO.HIGH)
    :param timeout: タイムアウト(default:0.25sec)
    :return: パルスの長さ(sec) タイムアウト時は0
    """

    start_time = time.time()
    pin_inv_value = not pin_value

    # 前のパルスが終了するのを待つ
    while GPIO.input(echo_pin) == pin_value:
        if time.time() - start_time > timeout:
            return RETURN_VALUE_ERROR

    # パルスの開始を待ち、時刻を記録する
    while GPIO.input(echo_pin) == pin_inv_value:
        if time.time() - start_time > timeout:
            return RETURN_VALUE_ERROR
    
    pulse_start_time = time.time()

    # パルスが終了するのを待ち、時刻を記録する
    while GPIO.input(echo_pin) == pin_value:
        if time.time() - start_time > timeout:
            return RETURN_VALUE_ERROR

    pulse_end_time = time.time()

    return pulse_end_time - pulse_start_time


def get_distance_hcsr04(trig_pin=26, echo_pin=19, temp=25):
    """
    超音波センサHC-SR04から距離を取得する
    :param trig_pin: TRIGのGPIO 番号(default:GPIO26)
    :param echo_pin: ECHOのGPIO 番号(default:GPIO19)
    :param temp: 音波速度補正用の温度(default:25)
    :return: 距離(m) タイムアウト時は0
    """

    # 出力を初期化
    GPIO.output(trig_pin, GPIO.LOW)
    time.sleep(0.3)

    # 出力(10us以上待つ)
    GPIO.output(trig_pin, GPIO.HIGH)
    time.sleep(0.00001)

    # 出力停止
    GPIO.output(trig_pin, GPIO.LOW)

    # echoから
    echo_time = measure_pulse_time(echo_pin, GPIO.HIGH)
    echo_dinstance = (331.5 + 0.61 * temp) * echo_time / 2

    return echo_dinstance


def main():
    """
    ファイル単独で実行された場合のメイン処理
    :param: なし
    :return: なし
    """

    try:
        # GPIOの初期化
        init_hcsr04(GPIO_SONIC_TRIG, GPIO_SONIC_ECHO, GPIO.BCM)

        while True:
            # 距離の取得
            distance = get_distance_hcsr04(GPIO_SONIC_TRIG, GPIO_SONIC_ECHO)
            print('distance= %1.3f m' % distance)
                
    # 例外処理
    except KeyboardInterrupt:
        print('Keyboard Interruption')

    except Exception as e:
        print('Error: Exception in main')
        print(type(e))
        print(e.args)
        print(e)


if __name__ == '__main__':
    main()