#-------------------------------------------------------------------------------------#
# We-Bot Pi 制御用プログラム(Ver.0 2019/07/31)
#-------------------------------------------------------------------------------------#
# ソフトウェアを利用するにあたっての注意点（必ず確認して下さい）
# ・本ソフトウェアは(株)ケイエスワイがデモ用に作成したソフトです。
# ・本ソフトウェアの利用にあたっては、全て利用者の自己責任で行って下さい。
# ・本ソフトウェアの改変、流用は全て利用者の責任において、自由に行って構いません。
# ・本ソフトウェアを使用して発生したあらゆる損害（金銭的損害、物理的損害等）は
#   全て利用者の責任において負担し、作者はいかなる損害も補償しません。
# ・本ソフトウェアの配布等に関しては、全て配布者の責任において行い、
#   作者は一切の責任を負わないものとします。
# ・作者は本ソフトウェアに関しての問い合わせ等には一切ご対応致しません。

#-------------------------------------------------------------------------------------#
# ライブラリ
#-------------------------------------------------------------------------------------#
import serial
import time
import struct
import RPi.GPIO as GPIO

#-------------------------------------------------------------------------------------#
# 定数定義
#-------------------------------------------------------------------------------------#
RETURN_VALUE_SUCCESS = 0        # 戻り値 正常時
RETURN_VALUE_ERROR = -1         # 戻り値 エラー時

GPIO_RESETN = 4                 # モーター制御基板リセット GPIO4(7ピン)

UART_PORT = '/dev/ttyS0'        # UARTデバイス名
BAUD_RATE = 115200              # UARTボーレート
RECV_TIMEOUT = 1                # UART受信タイムアウト

CMD_LEN = 1                     # コマンドのバイト長
ID_LEN = 1                      # デバイスIDのバイト長
DATA_LEN = 2                    # データのバイト長
SUM_LEN = 1                     # チェックサムのバイト長
SUM_MASK = 0xFF                 # チェックサムのデータマスク
ID_DATA_FORMTA_SINGLE = '<BH'   # ID＋DATA 1個の場合のフォーマット
ID_DATA_FORMTA_DOUBLE = '<BHBH' # ID＋DATA 2個の場合のフォーマット
ID_FORMAT_SINGLE = 'B'          # IDのみ 1個の場合のフォーマット

WRITE_CMD = 0x57                # 'W'のASCIIコード
READ_CMD = 0x52                 # 'R'のASCIIコード

LEFT_MOTOR_ID = 1               # 左モータID
RIGHT_MOTOR_ID = 2              # 右モータID
ADC_LEFT_VR_ID = 3              # 左サーメットトリマID
ADC_RIGHT_VR_ID = 4             # 右サーメットトリマID
ADC_BATTERY_ID = 5              # バッテリー電圧ID

PWM_PARAM_MAX = 100.0           # PWM指定 前進最大値
PWM_PARAM_MIN = -100.0          # PWM指定 後退最大値

PWM_CMD_FACTOR = 10             # PWM 指示値の倍率(マイコン)
PWM_CMD_ZERO = 1000             # PWM ゼロ指示値(マイコン)
PWM_CMD_FORWARD_MAX = 2000      # PWM 前進最大指示(マイコン)
PWM_CMD_REVERSE_MAX = 0         # PWM 後退最大指示(マイコン)

ADC_RESOLUTION = 4096           # ADC分解能(マイコン)
ADC_REF_VOL = 3300              # ADC基準電圧(マイコン)
ADC_FACTOR_BATTERY = 4          # ADC バッテリー電圧 倍率
ADC_FACTOR_VR_LEFT = 1          # ADC 左サーメットトリマ 倍率
ADC_FACTOR_VR_RIGHT = 1         # ADC 左サーメットトリマ 倍率

#-------------------------------------------------------------------------------------#
# 関数
#-------------------------------------------------------------------------------------#
def reset_on_motor_controller():
    """
    モーター制御基板をリセットする
    :param: なし
    :return: なし
    """

    GPIO.setwarnings(False)
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(GPIO_RESETN, GPIO.OUT)
    GPIO.output(GPIO_RESETN, GPIO.LOW)
    time.sleep(0.5)


def reset_off_motor_controller():
    """
    モーター制御基板のリセットを解除する
    :param: なし
    :return: なし
    """

    GPIO.setwarnings(False)
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(GPIO_RESETN, GPIO.OUT)
    GPIO.output(GPIO_RESETN, GPIO.HIGH)
    time.sleep(0.5)


def motor_stop(portname=UART_PORT):
    """
    左右のモーターを停止する
    :param portname: UARTポート
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, PWM_CMD_ZERO, RIGHT_MOTOR_ID, PWM_CMD_ZERO)
    ret = send_cmd_write(portname, data)
    return ret


def motor_forward(portname=UART_PORT):
    """
    左右のモーターを前進させる(PWM DUTY 100%)
    :param portname: UARTポート
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, PWM_CMD_FORWARD_MAX, RIGHT_MOTOR_ID, PWM_CMD_FORWARD_MAX)
    ret = send_cmd_write(portname, data)
    return ret


def motor_reverse(portname=UART_PORT):
    """
    左右のモーターを後退させる(PWM DUTY 100%)
    :param portname: UARTポート
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, PWM_CMD_REVERSE_MAX, RIGHT_MOTOR_ID, PWM_CMD_REVERSE_MAX)
    ret = send_cmd_write(portname, data)
    return ret


def motor_spin_turn_right(portname=UART_PORT):
    """
    左モーターを前進、右モータを後退させ、右にスピンターンさせる(PWM DUTY 100%)
    :param portname: UARTポート
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, PWM_CMD_FORWARD_MAX, RIGHT_MOTOR_ID, PWM_CMD_REVERSE_MAX)
    ret = send_cmd_write(portname, data)
    return ret


def motor_spin_turn_left(portname=UART_PORT):
    """
    左モーターを逆転、右モータを前進させ、左にスピンターンさせる(PWM DUTY 100%)
    :param portname: UARTポート
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, PWM_CMD_REVERSE_MAX, RIGHT_MOTOR_ID, PWM_CMD_FORWARD_MAX)
    ret = send_cmd_write(portname, data)
    return ret


def motor_turn_right_forward(portname=UART_PORT):
    """
    左モーターを前進、右モータを停止させ、右に曲がる(PWM DUTY 100%)
    :param portname: UARTポート
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, PWM_CMD_FORWARD_MAX, RIGHT_MOTOR_ID, PWM_CMD_ZERO)
    ret = send_cmd_write(portname, data)
    return ret


def motor_turn_left_forward(portname=UART_PORT):
    """
    左モータを停止、右モーターを前進させ、左に曲がる(PWM DUTY 100%)
    :param portname: UARTポート
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, PWM_CMD_ZERO, RIGHT_MOTOR_ID, PWM_CMD_FORWARD_MAX)
    ret = send_cmd_write(portname, data)
    return ret


def motor_turn_left_reverse(portname=UART_PORT):
    """
    左モーターを停止、右モータを後退させ、バックしながら左後ろに曲がる(PWM DUTY 100%)
    :param portname: UARTポート
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, PWM_CMD_ZERO, RIGHT_MOTOR_ID, PWM_CMD_REVERSE_MAX)
    ret = send_cmd_write(portname, data)
    return ret


def motor_turn_right_reverse(portname=UART_PORT):
    """
    左モーターを後退、右モータを停止させ、バックしながら右後ろに曲がる(PWM DUTY 100%)
    :param portname: UARTポート
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, PWM_CMD_REVERSE_MAX, RIGHT_MOTOR_ID, PWM_CMD_ZERO)
    ret = send_cmd_write(portname, data)
    return ret


def motor_set_pwm(portname=UART_PORT, left_pwm=0, right_pwm=0):
    """
    左右のモーターをPWM DUTY比を指定で動かす
    :param portname: UARTポート
    :param left_pwm: 左モータPWM(100.0:前進100%～-100.0:後退100%, 0で停止, 小数点以下1桁まで)
    :param right_pwm: 右モータPWM(100.0:前進100%～-100.0:後退100%, 0で停止, 小数点以下1桁まで)
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    # DUTY比の範囲チェック
    if not((left_pwm>=PWM_PARAM_MIN and left_pwm<=PWM_PARAM_MAX) \
        or (right_pwm>=PWM_PARAM_MIN and right_pwm<=PWM_PARAM_MAX)):    
        return RETURN_VALUE_ERROR

    # 10倍して、オフセットさせてから整数にする
    left_motor_cmd_value = int(left_pwm * PWM_CMD_FACTOR + PWM_CMD_ZERO)
    right_motor_cmd_value = int(right_pwm * PWM_CMD_FACTOR + PWM_CMD_ZERO)

    # 送信データを作成し、送信
    data = struct.pack(ID_DATA_FORMTA_DOUBLE, LEFT_MOTOR_ID, left_motor_cmd_value, RIGHT_MOTOR_ID, right_motor_cmd_value)
    ret = send_cmd_write(portname, data)
    return ret


def read_vr_left_volgate(portname=UART_PORT):
    """
    左側のサーメットトリマの電圧を取得する
    :param portname: UARTポート
    :return: 電圧値[mV](float)、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_FORMAT_SINGLE, ADC_LEFT_VR_ID)
    ret = send_cmd_read(portname, data)
    if ret >= 0:
        bat_volgate = ret
        return bat_volgate
    else:
        return RETURN_VALUE_ERROR


def read_vr_right_volgate(portname=UART_PORT):
    """
    右側のサーメットトリマの電圧を取得する
    :param portname: UARTポート
    :return: 電圧値[mV](float)、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_FORMAT_SINGLE, ADC_RIGHT_VR_ID)
    ret = send_cmd_read(portname, data)
    if ret >= 0:
        bat_volgate = ret
        return bat_volgate
    else:
        return RETURN_VALUE_ERROR


def read_battery_volgate(portname=UART_PORT):
    """
    バッテリー電圧を取得する
    :param portname: UARTポート
    :return: 電圧値(float)(mV)、失敗時:-1
    """

    # 送信データを作成し、送信
    data = struct.pack(ID_FORMAT_SINGLE, ADC_BATTERY_ID)
    ret = send_cmd_read(portname, data)
    if ret[0] >= 0:
        bat_volgate = int(ADC_FACTOR_BATTERY * ret[0] * ADC_REF_VOL / ADC_RESOLUTION)
        return bat_volgate
    else:
        return None


def send_cmd_write(portname, data):
    """
    送信データに対して、チェックサム計算を行い、UARTに送信し、コマンド応答を確認する
    :param portname: UARTポート
    :param data: 送信データ(バイナリデータ)
    :return: 完了ステータス 成功時:0、失敗時:-1
    """

    try:
        # シリアルポートをオープン
        comport = serial.Serial(port=portname, baudrate=115200, timeout=RECV_TIMEOUT)

        # 送信バッファにデータがあったらバッファクリア
        if comport.out_waiting > 0:
            comport.reset_output_buffer()

        # 受信バッファのクリア
        comport.reset_input_buffer()

        # 送信
        data_len = len(data)
        cmd_format = '%dB' % (CMD_LEN)
        cmd_data = struct.pack(cmd_format, WRITE_CMD)

        sum_format = '%dB' % (SUM_LEN)
        sum_value = (sum(cmd_data) + sum(data)) & SUM_MASK
        sum_data = struct.pack(sum_format, sum_value)

        comport.write(cmd_data)
        comport.write(data)
        comport.write(sum_data)

        # 受信
        recv_data_count = int(data_len / (ID_LEN + DATA_LEN))
        recv_len = CMD_LEN + recv_data_count + SUM_LEN

        recv_buf = comport.read(recv_len)
        recv_cmd = recv_buf[0]

        # 受信データからのSUM計算
        sum_data =  sum(recv_buf[0:-SUM_LEN]) & SUM_MASK

        # 受信SUMの切り出し
        recv_buf_sum = 0
        for i in range(SUM_LEN):
            recv_buf_sum += int(recv_buf[-SUM_LEN+i]) * (2 ** (8*i))
        
        recv_buf_sum = recv_buf_sum & SUM_MASK

        # シリアルポートのクローズ
        comport.close()

        # 応答コマンドの確認
        if recv_cmd==WRITE_CMD and recv_buf_sum==sum_data:
            return RETURN_VALUE_SUCCESS
        else:       # 応答コマンドのエラー
            return RETURN_VALUE_ERROR

    # 例外処理
    except serial.SerialException:      # シリアルポート指定エラー
        print('Error: send_cmd_read Serial Port Error')
        return RETURN_VALUE_ERROR
    except serial.SerialTimeoutException:       # 通信タイムアウト
        print('Error: send_cmd_read Serial Timeout Error')
        return RETURN_VALUE_ERROR
    except Exception as e:
        print('send_cmd_write Exception')
        print(type(e))
        print(e.args)
        print(e)
        return RETURN_VALUE_ERROR


def send_cmd_read(portname, data):
    """
    送信データに対して、チェックサム計算を行い、UARTに送信し、コマンド応答からデータを取得する
    :param portname: UARTポート
    :param data: 送信データ(バイナリデータ)
    :return: 成功時:リードデータ(list)、失敗時:None
    """

    try:
        # シリアルポートをオープン
        comport = serial.Serial(port=portname, baudrate=115200, timeout=RECV_TIMEOUT)

        # 送信バッファにデータがあったらバッファクリア
        if comport.out_waiting > 0:
            comport.reset_output_buffer()

        # 受信バッファのクリア
        comport.reset_input_buffer()

        # 送信
        data_len = len(data)
        cmd_format = '%dB' % (CMD_LEN)
        cmd_data = struct.pack(cmd_format, READ_CMD)

        sum_format = '%dB' % (SUM_LEN)
        sum_value = (sum(cmd_data) + sum(data)) & SUM_MASK
        sum_data = struct.pack(sum_format, sum_value)

        comport.write(cmd_data)
        comport.write(data)
        comport.write(sum_data)

        # 受信
        recv_len = CMD_LEN + data_len * (ID_LEN + DATA_LEN) + SUM_LEN
        recv_buf = comport.read(recv_len)
        recv_cmd = recv_buf[0]

        # データの切り出し
        recv_data_list =  []
        for i in range(data_len):
            data_start = CMD_LEN + (ID_LEN + DATA_LEN) * i + ID_LEN
            recv_data = 0
            for j in range(DATA_LEN):
                recv_data += int(recv_buf[data_start+j]) * (2 ** (8*j))
            
            recv_data_list.append(recv_data)

        # 受信データからのSUM計算
        sum_data =  sum(recv_buf[0:recv_len-SUM_LEN]) & SUM_MASK

        # 受信SUMの切り出し
        recv_buf_sum = 0
        for i in range(SUM_LEN):
            recv_buf_sum += int(recv_buf[-SUM_LEN+i]) * (2 ** (8*i))
        
        recv_buf_sum = recv_buf_sum & SUM_MASK

        # シリアルポートのクローズ
        comport.close()

        # 応答コマンドの確認
        if recv_cmd==READ_CMD and recv_buf_sum==sum_data:
            return recv_data_list
        else:       # 応答コマンドのエラー
            return None

    # 例外処理
    except serial.SerialException:      # シリアルポート指定エラー
        print('Error: send_cmd_read Serial Port Error')
        return None
    except serial.SerialTimeoutException:       # 通信タイムアウト
        print('Error: send_cmd_read Serial Timeout Error')
        return None
    except Exception as e:
        print('send_cmd_read Exception')
        print(type(e))
        print(e.args)
        print(e)
        return RETURN_VALUE_ERROR
