#-------------------------------------------------------------------------------------#
# We-Bot Pi デモ走行用プログラム(Ver.1 2019/08/09)
#-------------------------------------------------------------------------------------#
# ソフトウェアを利用するにあたっての注意点（必ず確認して下さい）
# ・本ソフトウェアは(株)ケイエスワイがデモ用に作成したソフトです。
# ・本ソフトウェアの利用にあたっては、全て利用者の自己責任で行って下さい。
# ・本ソフトウェアの改変、流用は全て利用者の責任において、自由に行って構いません。
# ・本ソフトウェアを使用して発生したあらゆる損害（金銭的損害、物理的損害等）は
#   全て利用者の責任において負担し、作者はいかなる損害も補償しません。
# ・本ソフトウェアの配布等に関しては、全て配布者の責任において行い、
#   作者は一切の責任を負わないものとします。
# ・作者は本ソフトウェアに関しての問い合わせ等には一切ご対応致しません。

#-------------------------------------------------------------------------------------#
# ライブラリ
#-------------------------------------------------------------------------------------#
import time
import random
import we_bot_pi
import hc_sr04_pi

#-------------------------------------------------------------------------------------#
# 定数定義
#-------------------------------------------------------------------------------------#
RETURN_VALUE_SUCCESS = 0        # 戻り値 正常時
RETURN_VALUE_ERROR = -1         # 戻り値 エラー時

UART_PORT = '/dev/ttyS0'        # UARTデバイス
GPIO_SONIC_TRIG = 26            # 超音波センサTRIGピン GPIO番号(BCM)
GPIO_SONIC_ECHO = 19            # 超音波センサECHOピン GPIO番号(BCM)

CYCLE_TIME_SEC = 5              # 旋回周期
BACK_TIME_SEC = 0.5             # バック時間
TURN_TIME_SEC_MIN = 2.0         # 最小旋回時間
TURN_TIME_SEC_MAX = 5.0         # 最大旋回時間

DISTANCE_THRESHOLD = 0.1        # 距離センサしきい値
BATTERY_LOW_VOLTAGE = 5800      # バッテリー低下しきい値

#-------------------------------------------------------------------------------------#
# 関数
#-------------------------------------------------------------------------------------#
def go_backword(time_sec=BACK_TIME_SEC):
    """
    指定時間バックする
    :param time_sec: バックする時間(sec)
    :return: なし
    """

    we_bot_pi.motor_reverse(UART_PORT)
    time.sleep(time_sec)


def random_spin_turn(time_sec_min=TURN_TIME_SEC_MIN, time_sec_max=TURN_TIME_SEC_MAX, direction=0):
    """
    ランダム時間スピンターンする
    :param time_sec_min: 最小旋回時間(sec)
    :param time_sec_max: 最大旋回時間(sec)
    :direction: 最大旋回時間(sec)
    :return: なし
    """

    # スピンターン
    random_sec = random.uniform(time_sec_min, time_sec_max)
    if direction == 0:
        we_bot_pi.motor_spin_turn_left(UART_PORT) 
    elif direction == 1:
        print('motor_spin_turn_right')
        we_bot_pi.motor_spin_turn_right(UART_PORT) 

    time.sleep(random_sec)
    print('motor_stop')
    we_bot_pi.motor_stop(UART_PORT)


def check_battery_voltage(portname=UART_PORT, min_voltage=BATTERY_LOW_VOLTAGE):
    """
    バッテリー電圧がしきい値を下回っていないか確認する
    :param portname: UARTポート
    :param min_voltage: バッテリー低下しきい値(mv)
    :return: 正常:0、低下:-1
    """

    # バッテリー電圧のチェック
    volgate = we_bot_pi.read_battery_volgate(UART_PORT)
    print('Battery Volgage= %d mV' % volgate)
    if volgate < BATTERY_LOW_VOLTAGE:
        print('Error: Low Battery!')
        return RETURN_VALUE_ERROR
    else:
        return RETURN_VALUE_SUCCESS


def main():
    """
    デモ走行用メインプログラム
    定期的にランダムに旋回しながら、走行する
    :param: なし
    :return: なし
    """
    # モーター制御基板のリセット解除
    print('motor controller reset')
    we_bot_pi.reset_on_motor_controller()
    print('motor controller reset off')
    we_bot_pi.reset_off_motor_controller()

    try:
        # 停止
        print('motor_stop')
        we_bot_pi.motor_stop(UART_PORT)
        time.sleep(0.5)

        start_time = time.time()        # 直進開始時間を取得

        while True:
            # バッテリー電圧のチェック
            battery_status = check_battery_voltage(UART_PORT, BATTERY_LOW_VOLTAGE)

            if battery_status == RETURN_VALUE_ERROR:    # 低下時
                battery_status = check_battery_voltage(UART_PORT, BATTERY_LOW_VOLTAGE)  # 再度電圧チェック

                if battery_status == RETURN_VALUE_ERROR:    # 2回連続で低下していたらエラー終了
                    we_bot_pi.motor_stop(UART_PORT)         # モーター停止
                    we_bot_pi.reset_on_motor_controller()
                    exit(1)     # プログラム終了

            pass_time = time.time() - start_time    # 直進経過時間を確認

            if pass_time > CYCLE_TIME_SEC:      # 直進で一定時間経過したら向きを変える
                # 停止
                print('motor_stop')
                we_bot_pi.motor_stop(UART_PORT)
                time.sleep(0.5)

                # バック
                print('motor_reverse')
                go_backword(BACK_TIME_SEC)
                time.sleep(0.5)

                # スピンターン
                turn_direction = random.randint(0, 1)   # 旋回方向はランダム
                print('motor_spin_turn')
                random_spin_turn(direction=turn_direction)
                time.sleep(0.5)

                start_time = time.time()        # 直進開始時間を新規取得

            print('motor_forward')
            we_bot_pi.motor_forward(UART_PORT)

    # 例外処理
    except KeyboardInterrupt:       # Ctrl+Cで停止させた場合
        print('Keyboard Interruption')
        print('motor_stop')
        we_bot_pi.motor_stop(UART_PORT)
        we_bot_pi.reset_on_motor_controller()

    except Exception as e:
        print('Error: Exception in main')
        print('motor_stop')
        we_bot_pi.motor_stop(UART_PORT)
        we_bot_pi.reset_on_motor_controller()
        print(type(e))
        print(e.args)
        print(e)


if __name__ == '__main__':
    main()